// Copyright (C) 2025 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
// Qt-Security score:significant reason:trivial-impl-only

#include "qsvgeasinginterface_p.h"

QT_BEGIN_NAMESPACE

QSvgEasingInterface::~QSvgEasingInterface()
    = default;

QT_END_NAMESPACE
