/****************************************************************************
** Meta object code from reading C++ file 'qsvgrenderer.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.12.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtsvg/src/svg/qsvgrenderer.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qsvgrenderer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.12.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN12QSvgRendererE_t {};
} // unnamed namespace

template <> constexpr inline auto QSvgRenderer::qt_create_metaobjectdata<qt_meta_tag_ZN12QSvgRendererE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QSvgRenderer",
        "repaintNeeded",
        "",
        "load",
        "filename",
        "contents",
        "QXmlStreamReader*",
        "render",
        "QPainter*",
        "p",
        "QRectF",
        "bounds",
        "elementId",
        "viewBox",
        "framesPerSecond",
        "currentFrame",
        "aspectRatioMode",
        "Qt::AspectRatioMode",
        "options",
        "QtSvg::Options",
        "animationEnabled"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'repaintNeeded'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'load'
        QtMocHelpers::SlotData<bool(const QString &)>(3, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { QMetaType::QString, 4 },
        }}),
        // Slot 'load'
        QtMocHelpers::SlotData<bool(const QByteArray &)>(3, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { QMetaType::QByteArray, 5 },
        }}),
        // Slot 'load'
        QtMocHelpers::SlotData<bool(QXmlStreamReader *)>(3, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 6, 5 },
        }}),
        // Slot 'render'
        QtMocHelpers::SlotData<void(QPainter *)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
        // Slot 'render'
        QtMocHelpers::SlotData<void(QPainter *, const QRectF &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { 0x80000000 | 10, 11 },
        }}),
        // Slot 'render'
        QtMocHelpers::SlotData<void(QPainter *, const QString &, const QRectF &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { QMetaType::QString, 12 }, { 0x80000000 | 10, 11 },
        }}),
        // Slot 'render'
        QtMocHelpers::SlotData<void(QPainter *, const QString &)>(7, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { QMetaType::QString, 12 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'viewBox'
        QtMocHelpers::PropertyData<QRectF>(13, 0x80000000 | 10, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet),
        // property 'framesPerSecond'
        QtMocHelpers::PropertyData<int>(14, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
        // property 'currentFrame'
        QtMocHelpers::PropertyData<int>(15, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
        // property 'aspectRatioMode'
        QtMocHelpers::PropertyData<Qt::AspectRatioMode>(16, 0x80000000 | 17, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet),
        // property 'options'
        QtMocHelpers::PropertyData<QtSvg::Options>(18, 0x80000000 | 19, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet),
        // property 'animationEnabled'
        QtMocHelpers::PropertyData<bool>(20, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QSvgRenderer, qt_meta_tag_ZN12QSvgRendererE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QSvgRenderer::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QSvgRendererE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QSvgRendererE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN12QSvgRendererE_t>.metaTypes,
    nullptr
} };

void QSvgRenderer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QSvgRenderer *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->repaintNeeded(); break;
        case 1: { bool _r = _t->load((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 2: { bool _r = _t->load((*reinterpret_cast<std::add_pointer_t<QByteArray>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 3: { bool _r = _t->load((*reinterpret_cast<std::add_pointer_t<QXmlStreamReader*>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 4: _t->render((*reinterpret_cast<std::add_pointer_t<QPainter*>>(_a[1]))); break;
        case 5: _t->render((*reinterpret_cast<std::add_pointer_t<QPainter*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QRectF>>(_a[2]))); break;
        case 6: _t->render((*reinterpret_cast<std::add_pointer_t<QPainter*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QRectF>>(_a[3]))); break;
        case 7: _t->render((*reinterpret_cast<std::add_pointer_t<QPainter*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QSvgRenderer::*)()>(_a, &QSvgRenderer::repaintNeeded, 0))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QRectF*>(_v) = _t->viewBoxF(); break;
        case 1: *reinterpret_cast<int*>(_v) = _t->framesPerSecond(); break;
        case 2: *reinterpret_cast<int*>(_v) = _t->currentFrame(); break;
        case 3: *reinterpret_cast<Qt::AspectRatioMode*>(_v) = _t->aspectRatioMode(); break;
        case 4: *reinterpret_cast<QtSvg::Options*>(_v) = _t->options(); break;
        case 5: *reinterpret_cast<bool*>(_v) = _t->isAnimationEnabled(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setViewBox(*reinterpret_cast<QRectF*>(_v)); break;
        case 1: _t->setFramesPerSecond(*reinterpret_cast<int*>(_v)); break;
        case 2: _t->setCurrentFrame(*reinterpret_cast<int*>(_v)); break;
        case 3: _t->setAspectRatioMode(*reinterpret_cast<Qt::AspectRatioMode*>(_v)); break;
        case 4: _t->setOptions(*reinterpret_cast<QtSvg::Options*>(_v)); break;
        case 5: _t->setAnimationEnabled(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *QSvgRenderer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QSvgRenderer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QSvgRendererE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QSvgRenderer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 8;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    return _id;
}

// SIGNAL 0
void QSvgRenderer::repaintNeeded()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}
QT_WARNING_POP
